from __future__ import annotations

import re

import pytest

from xdsl.dialects.arith import ConstantOp
from xdsl.dialects.builtin import DenseArrayBase, StringAttr, i32
from xdsl.dialects.test import TestTermOp
from xdsl.ir import Block, Region
from xdsl.irdl import (
    AttrSizedOperandSegments,
    AttrSizedRegionSegments,
    AttrSizedResultSegments,
    AttrSizedSuccessorSegments,
    IRDLOperation,
    SameVariadicOperandSize,
    attr_def,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    opt_operand_def,
    opt_prop_def,
    opt_region_def,
    opt_result_def,
    opt_successor_def,
    prop_def,
    region_def,
    result_def,
    successor_def,
    traits_def,
    var_operand_def,
    var_region_def,
    var_result_def,
    var_successor_def,
)
from xdsl.traits import IsTerminator
from xdsl.utils.exceptions import VerifyException

################################################################################
#                                 Results                                      #
################################################################################


@irdl_op_definition
class ResultOp(IRDLOperation):
    name = "test.result_op"

    res = result_def(StringAttr)


def test_result_builder():
    op = ResultOp.build(result_types=[StringAttr("")])
    op.verify()
    assert op.result_types == (StringAttr(""),)


def test_result_builder_exception():
    with pytest.raises(ValueError):
        ResultOp.build()


@irdl_op_definition
class OptResultOp(IRDLOperation):
    name = "test.opt_result_op"

    res = opt_result_def(StringAttr)


def test_opt_result_builder():
    op1 = OptResultOp.build(result_types=[[StringAttr("")]])
    op2 = OptResultOp.build(result_types=[[]])
    op3 = OptResultOp.build(result_types=[None])
    op1.verify()
    op2.verify()
    op3.verify()
    assert op1.result_types == (StringAttr(""),)
    assert len(op2.results) == 0
    assert len(op3.results) == 0


def test_opt_result_builder_two_args():
    with pytest.raises(ValueError) as _:
        OptResultOp.build(result_types=[[StringAttr(""), StringAttr("")]])


@irdl_op_definition
class VarResultOp(IRDLOperation):
    name = "test.var_result_op"

    res = var_result_def(StringAttr)


def test_var_result_builder():
    op = VarResultOp.build(result_types=[[StringAttr("0"), StringAttr("1")]])
    op.verify()
    assert op.result_types == (
        StringAttr("0"),
        StringAttr("1"),
    )


@irdl_op_definition
class TwoVarResultOp(IRDLOperation):
    name = "test.two_var_result_op"

    res1 = var_result_def(StringAttr)
    res2 = var_result_def(StringAttr)
    irdl_options = [AttrSizedResultSegments()]


def test_two_var_result_builder():
    op = TwoVarResultOp.build(
        result_types=[
            [StringAttr("0"), StringAttr("1")],
            [StringAttr("2"), StringAttr("3")],
        ]
    )
    op.verify()
    assert op.result_types == (
        StringAttr("0"),
        StringAttr("1"),
        StringAttr("2"),
        StringAttr("3"),
    )

    assert op.attributes[
        AttrSizedResultSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 2])


def test_two_var_result_builder2():
    op = TwoVarResultOp.build(
        result_types=[
            [StringAttr("0")],
            [StringAttr("1"), StringAttr("2"), StringAttr("3")],
        ]
    )
    op.verify()
    assert op.result_types == (
        StringAttr("0"),
        StringAttr("1"),
        StringAttr("2"),
        StringAttr("3"),
    )
    assert op.attributes[
        AttrSizedResultSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [1, 3])


@irdl_op_definition
class MixedResultOp(IRDLOperation):
    name = "test.mixed"

    res1 = var_result_def(StringAttr)
    res2 = result_def(StringAttr)
    res3 = var_result_def(StringAttr)
    irdl_options = [AttrSizedResultSegments()]


def test_var_mixed_builder():
    op = MixedResultOp.build(
        result_types=[
            [StringAttr("0"), StringAttr("1")],
            StringAttr("2"),
            [StringAttr("3"), StringAttr("4")],
        ]
    )
    op.verify()
    assert op.result_types == (
        StringAttr("0"),
        StringAttr("1"),
        StringAttr("2"),
        StringAttr("3"),
        StringAttr("4"),
    )

    assert op.attributes[
        AttrSizedResultSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 1, 2])


################################################################################
#                                 Operands                                     #
################################################################################


@irdl_op_definition
class OperandOp(IRDLOperation):
    name = "test.operand_op"

    res = operand_def(StringAttr)


def test_operand_builder_operation():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = OperandOp.build(operands=[op1])
    op2.verify()
    assert tuple(op2.operands) == (op1.res,)


def test_operand_builder_value():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = OperandOp.build(operands=[op1.res])
    op2.verify()
    assert tuple(op2.operands) == (op1.res,)


def test_operand_builder_exception():
    with pytest.raises(ValueError):
        OperandOp.build()


@irdl_op_definition
class OptOperandOp(IRDLOperation):
    name = "test.opt_operand_op"

    res = opt_operand_def(StringAttr)


def test_opt_operand_builder():
    op = ResultOp.build(result_types=[StringAttr("0")])
    op1 = OptOperandOp.build(operands=[[op]])
    op2 = OptOperandOp.build(operands=[[]])
    op1.verify()
    op2.verify()
    assert [operand for operand in op1.operands] == [op.res]
    assert len(op2.operands) == 0


def test_opt_operand_builder_two_args():
    op = ResultOp.build(result_types=[StringAttr("0")])
    with pytest.raises(ValueError) as _:
        OptOperandOp.build(operands=[[op, op]])


@irdl_op_definition
class VarOperandOp(IRDLOperation):
    name = "test.var_operand_op"

    res = var_operand_def(StringAttr)


def test_var_operand_builder():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = VarOperandOp.build(operands=[[op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1.res, op1.res)


@irdl_op_definition
class TwoVarOperandOp(IRDLOperation):
    name = "test.two_var_operand_op"

    res1 = var_operand_def(StringAttr)
    res2 = var_operand_def(StringAttr)
    irdl_options = [AttrSizedOperandSegments()]


# Define a similar operation with the segment sizes as a property to test this case
@irdl_op_definition
class TwoVarOperandPropOp(IRDLOperation):
    name = "test.two_var_operand_op"

    res1 = var_operand_def(StringAttr)
    res2 = var_operand_def(StringAttr)
    irdl_options = [AttrSizedOperandSegments(as_property=True)]


def test_two_var_operand_builder():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = TwoVarOperandOp.build(operands=[[op1, op1], [op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1.res, op1.res, op1.res, op1.res)
    assert op2.attributes[
        AttrSizedOperandSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 2])


def test_two_var_operand_builder2():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = TwoVarOperandOp.build(operands=[[op1], [op1, op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1.res, op1.res, op1.res, op1.res)
    assert op2.attributes[
        AttrSizedOperandSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [1, 3])


def test_two_var_operand_prop_builder():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = TwoVarOperandPropOp.build(operands=[[op1, op1], [op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1.res, op1.res, op1.res, op1.res)
    assert op2.properties[
        AttrSizedOperandSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 2])


def test_two_var_operand_prop_builder2():
    op1 = ResultOp.build(result_types=[StringAttr("0")])
    op2 = TwoVarOperandPropOp.build(operands=[[op1], [op1, op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1.res, op1.res, op1.res, op1.res)
    assert op2.properties[
        AttrSizedOperandSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [1, 3])


@irdl_op_definition
class SameSizeVarOperandOp(IRDLOperation):
    name = "test.same_size_var_operand_op"

    var1 = var_operand_def()
    op1 = operand_def()
    var2 = var_operand_def()
    irdl_options = [SameVariadicOperandSize()]


def test_same_size_operand_builder():
    op1 = ResultOp.build(result_types=[StringAttr("0")]).res
    op2 = SameSizeVarOperandOp.build(operands=[[op1, op1], op1, [op1, op1]])
    op2.verify()
    assert tuple(op2.operands) == (op1, op1, op1, op1, op1)
    op2 = SameSizeVarOperandOp.create(operands=[op1, op1, op1, op1, op1])
    op2.verify()
    assert (op2.var1, op2.op1, op2.var2) == ((op1, op1), op1, (op1, op1))


def test_same_size_operand_builder2():
    op1 = ResultOp.build(result_types=[StringAttr("0")]).res
    with pytest.raises(
        ValueError, match=re.escape("Variadic operands have different sizes: [1, 3]")
    ):
        SameSizeVarOperandOp.build(operands=[[op1], op1, [op1, op1, op1]])
    op2 = SameSizeVarOperandOp.create(operands=[op1, op1, op1, op1])
    with pytest.raises(
        VerifyException,
        match=re.escape(
            "Operation does not verify: Operation has 3 operands for 2 variadic operands marked as having the same size."
        ),
    ):
        op2.verify()


################################################################################
#                                Attribute                                     #
################################################################################


@irdl_op_definition
class AttrOp(IRDLOperation):
    name = "test.attr_op"
    attr = attr_def(StringAttr)


def test_attr_op():
    op = AttrOp.build(attributes={"attr": StringAttr("0")})
    op.verify()
    assert op.attr == StringAttr("0")


def test_attr_new_attr_op():
    op = AttrOp.build(attributes={"attr": StringAttr("0"), "new_attr": StringAttr("1")})
    op.verify()
    assert op.attr == StringAttr("0")
    assert op.attributes["new_attr"] == StringAttr("1")


@irdl_op_definition
class OptionalAttrOp(IRDLOperation):
    name = "test.opt_attr_op"

    opt_attr = opt_attr_def(StringAttr)


def test_optional_attr_op_empty():
    op = OptionalAttrOp.build()
    op.verify()
    assert op.opt_attr is None


################################################################################
#                                 Property                                     #
################################################################################


@irdl_op_definition
class PropertyOp(IRDLOperation):
    name = "test.prop_op"
    attr = prop_def(StringAttr)


def test_prop_op():
    op = PropertyOp.build(properties={"attr": StringAttr("0")})
    op.verify()
    assert op.attr == StringAttr("0")


@irdl_op_definition
class OptionalPropertyOp(IRDLOperation):
    name = "test.opt_prop_op"

    opt_attr = opt_prop_def(StringAttr)


def test_optional_prop_op_empty():
    op = OptionalPropertyOp.build()
    op.verify()
    assert op.opt_attr is None


################################################################################
#                                  Region                                      #
################################################################################


@irdl_op_definition
class RegionOp(IRDLOperation):
    name = "test.region_op"

    region = region_def()


def test_region_op_region():
    op = RegionOp.build(regions=[Region()])
    op.verify()

    assert not op.region.blocks


def test_region_op_blocks():
    op = RegionOp.build(regions=[[Block(), Block()]])
    op.verify()
    assert len(op.region.blocks) == 2


def test_region_op_ops():
    op1 = RegionOp.build(regions=[[Block([TestTermOp.create()])]])
    op2 = RegionOp.build(regions=[[Block([TestTermOp.create()])]])
    op = RegionOp.build(regions=[[op1, op2, TestTermOp.create()]])
    op.verify()
    assert len(op.region.blocks) == 1
    assert len(op.region.blocks[0].ops) == 3


def test_noop_region():
    region0 = Region(Block())
    assert len(region0.ops) == 0


def test_singleop_region():
    a = ConstantOp.from_int_and_width(1, i32)
    region0 = Region(Block([a]))
    assert type(region0.op) is ConstantOp


@irdl_op_definition
class SBRegionOp(IRDLOperation):
    name = "test.sbregion_op"

    region = region_def("single_block")


def test_sbregion_one_block():
    op = SBRegionOp.build(regions=[[Block([TestTermOp.create()])]])
    op.verify()
    assert len(op.region.blocks) == 1


@irdl_op_definition
class OptRegionOp(IRDLOperation):
    name = "test.opt_region_op"

    reg = opt_region_def()


def test_opt_region_builder():
    op1 = OptRegionOp.build(regions=[[[Block(), Block()]]])
    op2 = OptRegionOp.build(regions=[[Region()]])
    op3 = OptRegionOp.build(regions=[None])
    op1.verify()
    op2.verify()
    op3.verify()


def test_opt_region_builder_two_args():
    with pytest.raises(ValueError) as _:
        OptRegionOp.build(regions=[[Region(), Region()]])


@irdl_op_definition
class OptSBRegionOp(IRDLOperation):
    name = "test.sbregion_op"

    region = opt_region_def("single_block")


def test_opt_sbregion_one_block():
    op1 = OptSBRegionOp.build(regions=[[[Block([TestTermOp.create()])]]])
    op2 = OptSBRegionOp.build(regions=[[]])
    op1.verify()
    op2.verify()
    assert op1.region is not None
    assert len(op1.region.blocks) == 1
    assert op2.region is None


@irdl_op_definition
class VarRegionOp(IRDLOperation):
    name = "test.var_operand_op"

    regs = var_region_def()


def test_var_region_builder():
    op = VarRegionOp.build(regions=[[Region(), [Block(), Block()]]])
    op.verify()
    assert len(op.regs[0].blocks) == 0
    assert len(op.regs[1].blocks) == 2


@irdl_op_definition
class VarSBRegionOp(IRDLOperation):
    name = "test.sbregion_op"

    regs = var_region_def("single_block")


def test_var_sbregion_one_block():
    op1 = VarSBRegionOp.build(regions=[[[Block([TestTermOp.create()])]]])
    op2 = VarSBRegionOp.build(regions=[[Region(), [Block(), Block()]]])
    op1.verify()
    op2.verify()
    assert len(op1.regs) == 1
    assert len(op2.regs) == 2
    assert len(op2.regs[0].blocks) == 0
    assert len(op2.regs[1].blocks) == 2


@irdl_op_definition
class TwoVarRegionOp(IRDLOperation):
    name = "test.two_var_region_op"

    res1 = var_region_def()
    res2 = var_region_def()
    irdl_options = [AttrSizedRegionSegments()]


def test_two_var_region_builder():
    region1 = Region()
    region2 = Region()
    region3 = Region()
    region4 = Region()
    op2 = TwoVarRegionOp.build(regions=[[region1, region2], [region3, region4]])
    op2.verify()
    assert op2.regions == (region1, region2, region3, region4)
    assert op2.attributes[
        AttrSizedRegionSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 2])


def test_two_var_operand_builder3():
    region1 = Region()
    region2 = Region()
    region3 = Region()
    region4 = Region()
    op2 = TwoVarRegionOp.build(regions=[[region1], [region2, region3, region4]])
    op2.verify()
    assert op2.regions == (region1, region2, region3, region4)
    assert op2.attributes[
        AttrSizedRegionSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [1, 3])


################################################################################
#                                Successor                                     #
################################################################################


@irdl_op_definition
class SuccessorOp(IRDLOperation):
    name = "test.successor_op"

    successor = successor_def()

    traits = traits_def(IsTerminator())


def test_successor_op_successor():
    """Test operation from IRDL operation definition can have successors"""
    block0 = Block()

    op = SuccessorOp.build(successors=[block0])
    block1 = Block([op])

    _ = Region([block0, block1])

    op.verify()
    assert len(op.successors) == 1


@irdl_op_definition
class OptSuccessorOp(IRDLOperation):
    name = "test.opt_successora_op"

    successor = opt_successor_def()

    traits = traits_def(IsTerminator())


def test_opt_successor_builder():
    """
    Test operation from IRDL operation definition can optionally have
    successors
    """
    block0 = Block()

    op1 = OptSuccessorOp.build(successors=[block0])
    block1 = Block([op1])

    op2 = OptSuccessorOp.build(successors=[None])
    block2 = Block([op2])

    _ = Region([block0, block1, block2])

    op1.verify()
    op2.verify()


@irdl_op_definition
class VarSuccessorOp(IRDLOperation):
    name = "test.var_succesor_op"

    successor = var_successor_def()

    traits = traits_def(IsTerminator())


def test_var_successor_builder():
    """
    Test operation from IRDL operation definition can have variadic successors
    """
    block0 = Block()
    op = VarSuccessorOp.build(successors=[[block0, block0, block0]])

    block1 = Block([op])
    _ = Region([block0, block1])

    op.verify()
    assert len(op.successors) == 3


@irdl_op_definition
class TwoVarSuccessorOp(IRDLOperation):
    name = "test.two_var_successor_op"

    res1 = var_successor_def()
    res2 = var_successor_def()
    irdl_options = [AttrSizedSuccessorSegments()]

    traits = traits_def(IsTerminator())


def test_two_var_successor_builder():
    """
    Test operation from IRDL operation definition can have variadic successors
    along with their sizes as an attribute
    """
    block1 = Block()
    block2 = Block()
    block3 = Block()
    block4 = Block()

    op2 = TwoVarSuccessorOp.build(successors=[[block1, block2], [block3, block4]])
    block0 = Block([op2])

    _ = Region([block0, block1, block2, block3, block4])

    op2.verify()
    assert tuple(op2.successors) == (block1, block2, block3, block4)
    assert op2.attributes[
        AttrSizedSuccessorSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [2, 2])


def test_two_var_successor_builder2():
    """
    Test operation from IRDL operation definition can have variadic successors
    along with their sizes as an attribute
    """
    block1 = Block()
    block2 = Block()
    block3 = Block()
    block4 = Block()

    op2 = TwoVarSuccessorOp.build(successors=[[block1], [block2, block3, block4]])
    block0 = Block([op2])

    _ = Region([block0, block1, block2, block3, block4])

    op2.verify()
    assert tuple(op2.successors) == (block1, block2, block3, block4)
    assert op2.attributes[
        AttrSizedSuccessorSegments.attribute_name
    ] == DenseArrayBase.from_list(i32, [1, 3])


################################################################################
#                                   Misc                                       #
################################################################################


def test_parent_pointers():
    op = ResultOp.build(result_types=[StringAttr("0")])
    block = Block([op])
    reg = Region(block)
    reg_op = RegionOp.build(regions=[reg])

    block_2 = Block([reg_op])
    reg_2 = Region(block_2)
    reg_op_2 = RegionOp.build(regions=[reg_2])

    assert op.parent_block() is block
    assert op.parent_region() is reg
    assert op.parent_op() is reg_op

    assert reg_op_2.parent_block() is None
    assert reg_op_2.parent_region() is None
    assert reg_op_2.parent_op() is None

    assert reg.parent_op() is reg_op
    assert reg.parent_block() is block_2
    assert reg.parent_region() is reg_2

    assert reg_2.parent_block() is None
    assert reg_2.parent_region() is None

    assert block.parent_region() is reg
    assert block.parent_op() is reg_op
    assert block.parent_block() is block_2

    assert block_2.parent_block() is None



import abc
import json
from collections.abc import Iterable, Sequence
from itertools import chain
from typing import cast

from ordered_set import OrderedSet

from xdsl.backend.register_queue import RegisterQueue
from xdsl.dialects import riscv, riscv_func, riscv_scf, riscv_snitch
from xdsl.dialects.riscv import (
    FloatRegisterType,
    IntRegisterType,
    Registers,
    RISCVAsmOperation,
    RISCVRegisterType,
)
from xdsl.ir import Attribute, Block, Operation, SSAValue
from xdsl.rewriter import InsertPoint, Rewriter
from xdsl.transforms.canonicalization_patterns.riscv import get_constant_value
from xdsl.transforms.snitch_register_allocation import get_snitch_reserved
from xdsl.utils.exceptions import DiagnosticException


def gather_allocated(
    func: riscv_func.FuncOp,
) -> set[IntRegisterType | FloatRegisterType]:
    """Utility method to gather already allocated registers"""

    allocated: set[IntRegisterType | FloatRegisterType] = set()

    for op in func.walk():
        if not isinstance(op, RISCVAsmOperation):
            continue

        if isinstance(op, riscv_func.CallOp):
            # These registers are not guaranteed to hold the same values when the callee
            # returns, according to the RISC-V calling convention.
            # https://riscv.org/wp-content/uploads/2015/01/riscv-calling.pdf
            allocated.update(riscv.Registers.A)
            allocated.update(riscv.Registers.T)
            allocated.update(riscv.Registers.FA)
            allocated.update(riscv.Registers.FT)

        for param in chain(op.operands, op.results):
            if (
                isinstance(param.type, IntRegisterType | FloatRegisterType)
                and param.type.is_allocated
            ):
                if not param.type.register_name.data.startswith("j"):
                    allocated.add(param.type)

    return allocated


def _uses_snitch_stream(func: riscv_func.FuncOp) -> bool:
    """Utility method to detect use of read/write ops of the `snitch_stream` dialect."""

    return any(
        isinstance(op, riscv_snitch.ReadOp | riscv_snitch.WriteOp) for op in func.walk()
    )


class RegisterAllocator(abc.ABC):
    """
    Base class for register allocation strategies.
    """

    @abc.abstractmethod
    def allocate_func(self, func: riscv_func.FuncOp) -> None:
        raise NotImplementedError()


def reg_types(regs: Iterable[Attribute]) -> tuple[set[str], set[str]]:
    """
    Returns a tuple containing the sets of IntRegister and FloatRegister in the iterable.
    """
    int_regs: set[str] = set()
    float_regs: set[str] = set()

    for reg in regs:
        if isinstance(reg, IntRegisterType):
            int_regs.add(reg.register_name.data)
        elif isinstance(reg, FloatRegisterType):
            float_regs.add(reg.register_name.data)

    return int_regs, float_regs


class RegisterAllocatorLivenessBlockNaive(RegisterAllocator):
    """
    It traverses the use-def SSA chain backwards (i.e., from uses to defs) and:
      1. allocates registers for operands
      2. frees registers for results (since that will be the last time they appear when
      going backwards)
    It currently operates on blocks.

    This is a simplified version of the standard bottom-up local register allocation
    algorithm.

    A relevant reference in "Engineering a Compiler, Edition 3" ISBN: 9780128154120.

    ```
    for op in block.walk_reverse():
    for operand in op.operands:
        if operand is not allocated:
            allocate(operand)

    for result in op.results:
    if result is not allocated:
        allocate(result)
        free_before_next_instruction.append(result)
    else:
        free(result)
    ```
    """

    available_registers: RegisterQueue[IntRegisterType | FloatRegisterType]
    live_ins_per_block: dict[Block, OrderedSet[SSAValue]]

    exclude_preallocated: bool = True
    exclude_snitch_reserved: bool = True

    def __init__(
        self, available_registers: RegisterQueue[IntRegisterType | FloatRegisterType]
    ) -> None:
        self.available_registers = available_registers
        self.live_ins_per_block = {}

    def allocate(self, reg: SSAValue) -> bool:
        """
        Allocate a register if not already allocated.
        """
        if (
            isinstance(reg.type, IntRegisterType | FloatRegisterType)
            and not reg.type.is_allocated
        ):
            if (val := get_constant_value(reg)) is not None and val.value.data == 0:
                reg.type = Registers.ZERO
            else:
                reg.type = self.available_registers.pop(type(reg.type))
            return True

        return False

    def allocate_same(self, vals: Sequence[SSAValue]) -> bool:
        """
        Allocates the values passed in to the same register.
        If some of the values are already allocated, they must be allocated to the same
        register, and unallocated values are then allocated to this register.
        If the values passed in are already allocated to differing registers, a
        `DiagnosticException` is raised.
        """
        reg_types = set(val.type for val in vals)
        assert all(isinstance(reg_type, RISCVRegisterType) for reg_type in reg_types)
        reg_types = cast(set[IntRegisterType | FloatRegisterType], reg_types)

        match len(reg_types):
            case 0:
                # No inputs, nothing to do
                return False
            case 1:
                # Single input, may already be allocated
                reg_type = next(iter(reg_types))
                if reg_type.is_allocated:
                    return False
                else:
                    reg_type = self.available_registers.pop(type(reg_type))
            case 2:
                # Two inputs, either one is allocated or two
                reg_type_0, reg_type_1 = reg_types
                if reg_type_0.is_allocated:
                    if reg_type_1.is_allocated:
                        reg_names = [f"{reg_type}" for reg_type in reg_types]
                        reg_names.sort()
                        raise DiagnosticException(
                            f"Cannot allocate registers to the same register {reg_names}"
                        )
                    else:
                        reg_type = reg_type_0
                else:
                    reg_type = reg_type_1
            case _:
                # More than one input is allocated, meaning we can't allocate them to be
                # the same, error.
                reg_names = [f"{reg_type}" for reg_type in reg_types]
                reg_names.sort()
                raise DiagnosticException(
                    f"Cannot allocate registers to the same register {reg_names}"
                )

        did_allocate = False

        for val in vals:
            if val.type != reg_type:
                val.type = reg_type
                did_allocate = True

        return did_allocate

    def _free(self, reg: SSAValue) -> None:
        if (
            isinstance(reg.type, IntRegisterType | FloatRegisterType)
            and reg.type.is_allocated
        ):
            self.available_registers.push(reg.type)

    def process_operation(self, op: Operation) -> None:
        """
        Allocate registers for one operation.
        """
        match op:
            case riscv_scf.ForOp():
                self.allocate_for_loop(op)
            case riscv_snitch.FRepOperation():
                self.allocate_frep_loop(op)
            case RISCVAsmOperation():
                self.process_riscv_op(op)
            case _:
                # Ignore non-riscv operations
                return

    def process_riscv_op(self, op: RISCVAsmOperation) -> None:
        """
        Allocate registers for RISC-V Instruction.
        """
        ins, outs, inouts = op.get_register_constraints()

        # Allocate registers to inout operand groups since they are defined further up
        # in the use-def SSA chain
        for operand_group in inouts:
            self.allocate_same(operand_group)

        for result in outs:
            # Allocate registers to result if not already allocated
            self.allocate(result)
            # Free the register since the SSA value is created here
            self._free(result)

        # Allocate registers to operands since they are defined further up
        # in the use-def SSA chain
        for operand in ins:
            self.allocate(operand)

    def allocate_for_loop(self, loop: riscv_scf.ForOp) -> None:
        """
        Allocate registers for riscv_scf for loop, recursively calling process_operation
        for operations in the loop.
        """
        # Allocate values used inside the body but defined outside.
        # Their scope lasts for the whole body execution scope
        live_ins = self.live_ins_per_block[loop.body.block]
        for live_in in live_ins:
            self.allocate(live_in)

        yield_op = loop.body.block.last_op
        assert yield_op is not None, (
            "last op of riscv_scf.ForOp is guaranteed to be riscv_scf.Yield"
        )
        block_args = loop.body.block.args

        # The loop-carried variables are trickier
        # The for op operand, block arg, and yield operand must have the same type
        for block_arg, operand, yield_operand, op_result in zip(
            block_args[1:], loop.iter_args, yield_op.operands, loop.results
        ):
            self.allocate_same((block_arg, operand, yield_operand, op_result))

        # Induction variable
        assert isinstance(block_args[0].type, IntRegisterType)
        self.allocate(block_args[0])

        # Step and ub are used throughout loop
        self.allocate(loop.ub)
        self.allocate(loop.step)

        # Reserve the loop carried variables for allocation within the body
        regs = loop.iter_args.types
        assert all(isinstance(reg, IntRegisterType | FloatRegisterType) for reg in regs)
        regs = cast(tuple[IntRegisterType | FloatRegisterType], regs)
        with self.available_registers.reserve_registers(regs):
            for op in reversed(loop.body.block.ops):
                self.process_operation(op)

        # lb is only used as an input to the loop, so free induction variable before
        # allocating lb to it in case it's not yet allocated
        self._free(block_args[0])
        self.allocate(loop.lb)

    def allocate_frep_loop(self, loop: riscv_snitch.FRepOperation) -> None:
        """
        Allocate registers for riscv_snitch frep_outer or frep_inner loop, recursively
        calling process_operation for operations in the loop.
        """
        # Allocate values used inside the body but defined outside.
        # Their scope lasts for the whole body execution scope
        live_ins = self.live_ins_per_block[loop.body.block]
        for live_in in live_ins:
            self.allocate(live_in)

        yield_op = loop.body.block.last_op
        assert yield_op is not None, (
            "last op of riscv_snitch.frep_outer and riscv_snitch.frep_inner is guaranteed"
            " to be riscv_scf.Yield"
        )
        block_args = loop.body.block.args

        # The loop-carried variables are trickier
        # The for op operand, block arg, and yield operand must have the same type
        for block_arg, operand, yield_operand, op_result in zip(
            block_args, loop.iter_args, yield_op.operands, loop.results
        ):
            self.allocate_same((block_arg, operand, yield_operand, op_result))

        self.allocate(loop.max_rep)

        # Reserve the loop carried variables for allocation within the body
        regs = loop.iter_args.types
        assert all(isinstance(reg, IntRegisterType | FloatRegisterType) for reg in regs)
        regs = cast(tuple[IntRegisterType | FloatRegisterType], regs)
        with self.available_registers.reserve_registers(regs):
            for op in reversed(loop.body.block.ops):
                self.process_operation(op)

    def allocate_func(
        self, func: riscv_func.FuncOp, *, add_regalloc_stats: bool = False
    ) -> None:
        """
        Allocates values in function passed in to registers.
        The whole function must have been lowered to the relevant riscv dialects
        and it must contain no unrealized casts.
        If `add_regalloc_stats` is set to `True`, then a comment op will be inserted
        before the function op passed in with a json containing the relevant data.
        """
        if not func.body.blocks:
            # External function declaration
            return

        if len(func.body.blocks) != 1:
            raise NotImplementedError(
                f"Cannot register allocate func with {len(func.body.blocks)} blocks."
            )

        preallocated: set[IntRegisterType | FloatRegisterType] = set()

        if self.exclude_preallocated:
            preallocated |= gather_allocated(func)

        if self.exclude_snitch_reserved and _uses_snitch_stream(func):
            preallocated |= get_snitch_reserved()

        for pa_reg in preallocated:
            self.available_registers.reserve_register(pa_reg)
            self.available_registers.exclude_register(pa_reg)

        block = func.body.block

        self.live_ins_per_block = live_ins_per_block(block)
        assert not self.live_ins_per_block[block]
        for op in reversed(block.ops):
            self.process_operation(op)

        if add_regalloc_stats:
            preallocated_int, preallocated_float = reg_types(preallocated)
            allocated_int, allocated_float = reg_types(
                val.type
                for op in block.walk()
                for vals in (op.results, op.operands)
                for val in vals
            )

            stats = {
                "preallocated_float": sorted(preallocated_float),
                "preallocated_int": sorted(preallocated_int),
                "allocated_float": sorted(allocated_float),
                "allocated_int": sorted(allocated_int),
            }

            stats_str = json.dumps(stats)

            Rewriter.insert_op(
                riscv.CommentOp(f"Regalloc stats: {stats_str}"),
                InsertPoint.before(func),
            )


def _live_ins_per_block(
    block: Block, acc: dict[Block, OrderedSet[SSAValue]]
) -> OrderedSet[SSAValue]:
    res = OrderedSet[SSAValue]([])

    for op in reversed(block.ops):
        # Remove values defined in the block
        # We are traversing backwards, so cannot use the value removed here again
        res.difference_update(op.results)
        # Add values used in the block
        res.update(op.operands)

        # Process inner blocks
        for region in op.regions:
            for inner in region.blocks:
                # Add the values used in the inner block
                res.update(_live_ins_per_block(inner, acc))

    # Remove the block arguments
    res.difference_update(block.args)

    acc[block] = res

    return res


def live_ins_per_block(block: Block) -> dict[Block, OrderedSet[SSAValue]]:
    """
    Returns a mapping from a block to the set of values used in it but defined outside of
    it.
    """
    res: dict[Block, OrderedSet[SSAValue]] = {}
    _ = _live_ins_per_block(block, res)
    return res



import pytest

from xdsl.context import Context
from xdsl.dialects.builtin import UnregisteredAttr, UnregisteredOp
from xdsl.ir import Dialect, ParametrizedAttribute, TypeAttribute
from xdsl.irdl import IRDLOperation, irdl_attr_definition, irdl_op_definition


@irdl_op_definition
class DummyOp(IRDLOperation):
    name = "test.dummy"


@irdl_op_definition
class DummyOp2(IRDLOperation):
    name = "test.dummy2"


@irdl_attr_definition
class DummyAttr(ParametrizedAttribute):
    name = "test.dummy_attr"


@irdl_attr_definition
class DummyAttr2(ParametrizedAttribute):
    name = "test.dummy_attr2"


testDialect = Dialect("test", [DummyOp], [DummyAttr])
testDialect2 = Dialect("test", [DummyOp2], [DummyAttr2])


def test_get_op():
    """Test `get_op` and `get_optional_op` methods."""
    ctx = Context()
    ctx.load_op(DummyOp)

    assert ctx.get_op("test.dummy") == DummyOp
    with pytest.raises(Exception):
        _ = ctx.get_op("test.dummy2")

    assert ctx.get_optional_op("test.dummy") == DummyOp
    assert ctx.get_optional_op("test.dummy2") is None


def test_get_op_unregistered():
    """
    Test `get_op` and `get_optional_op`
    methods with the `allow_unregistered` flag.
    """
    ctx = Context(allow_unregistered=True)
    ctx.load_op(DummyOp)

    assert ctx.get_optional_op("test.dummy") == DummyOp
    op = ctx.get_optional_op("test.dummy2")
    assert op is not None
    assert issubclass(op, UnregisteredOp)

    assert ctx.get_op("test.dummy") == DummyOp
    assert issubclass(ctx.get_op("test.dummy2"), UnregisteredOp)


def test_get_op_with_dialect_stack():
    """Test `get_op` and `get_optional_op` methods."""
    ctx = Context()
    ctx.load_op(DummyOp)

    assert ctx.get_op("dummy", dialect_stack=("test",)) == DummyOp
    with pytest.raises(Exception):
        _ = ctx.get_op("dummy2", dialect_stack=("test",))

    assert ctx.get_optional_op("dummy", dialect_stack=("test",)) == DummyOp
    assert ctx.get_optional_op("dummy2", dialect_stack=("test",)) is None


def test_get_op_unregistered_with_dialect_stack():
    """
    Test `get_op` and `get_optional_op`
    methods with the `allow_unregistered` flag.
    """
    ctx = Context(allow_unregistered=True)
    ctx.load_op(DummyOp)

    assert ctx.get_optional_op("dummy", dialect_stack=("test",)) == DummyOp
    op_type = ctx.get_optional_op("dummy2", dialect_stack=("test",))
    assert op_type is not None
    assert issubclass(op_type, UnregisteredOp)
    assert op_type.create().op_name.data == "dummy2"

    assert ctx.get_op("dummy", dialect_stack=("test",)) == DummyOp
    op_type = ctx.get_op("dummy2", dialect_stack=("test",))
    assert issubclass(op_type, UnregisteredOp)
    assert op_type.create().op_name.data == "dummy2"


def test_get_attr():
    """Test `get_attr` and `get_optional_attr` methods."""
    ctx = Context()
    ctx.load_attr(DummyAttr)

    assert ctx.get_attr("test.dummy_attr") == DummyAttr
    with pytest.raises(Exception):
        _ = ctx.get_attr("test.dummy_attr2")

    assert ctx.get_optional_attr("test.dummy_attr") == DummyAttr
    assert ctx.get_optional_attr("test.dummy_attr2") is None


@pytest.mark.parametrize("is_type", [True, False])
def test_get_attr_unregistered(is_type: bool):
    """
    Test `get_attr` and `get_optional_attr`
    methods with the `allow_unregistered` flag.
    """
    ctx = Context(allow_unregistered=True)
    ctx.load_attr(DummyAttr)

    assert (
        ctx.get_optional_attr("test.dummy_attr", create_unregistered_as_type=is_type)
        == DummyAttr
    )
    attr = ctx.get_optional_attr("test.dummy_attr2")
    assert attr is not None
    assert issubclass(attr, UnregisteredAttr)
    if is_type:
        assert issubclass(attr, TypeAttribute)

    assert (
        ctx.get_attr("test.dummy_attr", create_unregistered_as_type=is_type)
        == DummyAttr
    )
    assert issubclass(
        ctx.get_attr("test.dummy_attr2", create_unregistered_as_type=is_type),
        UnregisteredAttr,
    )
    if is_type:
        assert issubclass(attr, TypeAttribute)


def test_clone_function():
    ctx = Context()
    ctx.load_attr(DummyAttr)
    ctx.load_op(DummyOp)

    copy = ctx.clone()

    assert ctx == copy
    copy.load_op(DummyOp2)
    assert ctx != copy

    copy = ctx.clone()

    assert ctx == copy
    copy.load_attr(DummyAttr2)
    assert ctx != copy


def test_register_dialect_get_op_attr():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    op = ctx.get_op("test.dummy")
    assert op == DummyOp
    attr = ctx.get_attr("test.dummy_attr")
    assert attr == DummyAttr
    assert "test" in ctx.registered_dialect_names
    assert list(ctx.loaded_dialects) == [testDialect]


def test_register_dialect_already_registered():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    with pytest.raises(ValueError, match="'test' dialect is already registered"):
        ctx.register_dialect("test", lambda: testDialect2)


def test_register_dialect_already_loaded():
    ctx = Context()
    ctx.load_dialect(testDialect)
    with pytest.raises(ValueError, match="'test' dialect is already registered"):
        ctx.register_dialect("test", lambda: testDialect2)


def test_load_registered_dialect():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    assert list(ctx.loaded_dialects) == []
    assert list(ctx.registered_dialect_names) == ["test"]
    ctx.load_registered_dialect("test")
    assert list(ctx.loaded_dialects) == [testDialect]
    assert list(ctx.registered_dialect_names) == ["test"]


def test_load_registered_dialect_not_registered():
    ctx = Context()
    with pytest.raises(ValueError, match="'test' dialect is not registered"):
        ctx.load_registered_dialect("test")


def test_load_dialect():
    ctx = Context()
    ctx.load_dialect(testDialect)
    assert list(ctx.loaded_dialects) == [testDialect]
    assert list(ctx.registered_dialect_names) == ["test"]


def test_load_dialect_already_loaded():
    ctx = Context()
    ctx.load_dialect(testDialect)
    with pytest.raises(ValueError, match="'test' dialect is already registered"):
        ctx.load_dialect(testDialect)


def test_load_dialect_already_registered():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    with pytest.raises(
        ValueError,
        match="'test' dialect is already registered, use "
        "'load_registered_dialect' instead",
    ):
        ctx.load_dialect(testDialect)


def test_get_optional_op_registered():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    assert ctx.get_optional_op("test.dummy") == DummyOp
    assert ctx.get_optional_op("test.dummy2") is None
    assert list(ctx.loaded_dialects) == [testDialect]
    assert list(ctx.registered_dialect_names) == ["test"]


def test_get_optional_attr_registered():
    ctx = Context()
    ctx.register_dialect("test", lambda: testDialect)
    assert ctx.get_optional_attr("test.dummy_attr") == DummyAttr
    assert ctx.get_optional_attr("test.dummy_attr2") is None
    assert list(ctx.loaded_dialects) == [testDialect]
    assert list(ctx.registered_dialect_names) == ["test"]



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "xdsl==0.27.0",
# ]
# ///

import marimo

__generated_with = "0.11.10"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo

    from sympy import (
        S,
        symbols,
        Expr,
        Add,
        Mul,
        Sum,
        Integer,
        Float,
        E,
        I,
        re,
        im,
        Abs,
        Pow,
        Rational,
        Function,
    )
    from sympy.core.symbol import Symbol

    from xdsl.ir import (
        Operation,
        SSAValue,
        Region,
        Block,
        ParametrizedAttribute,
        Attribute,
        TypeAttribute,
    )
    from xdsl.pattern_rewriter import (
        PatternRewriter,
        RewritePattern,
        op_type_rewrite_pattern,
        PatternRewriteWalker,
        GreedyRewritePatternApplier,
    )
    from xdsl.transforms.dead_code_elimination import region_dce
    from xdsl.traits import Pure
    from xdsl.irdl import (
        irdl_op_definition,
        IRDLOperation,
        irdl_attr_definition,
        operand_def,
        result_def,
        ParameterDef,
        traits_def,
    )
    from xdsl.dialects.builtin import (
        ModuleOp,
        Float64Type,
        FloatAttr,
        IntegerType,
        IntegerAttr,
    )
    from xdsl.dialects.func import FuncOp, ReturnOp
    from xdsl.dialects.arith import (
        AddfOp,
        SubfOp,
        MulfOp,
        ConstantOp,
        AddiOp,
        MuliOp,
        SIToFPOp,
        FloatingPointLikeBinaryOperation,
        DivfOp,
    )
    from xdsl.dialects.scf import ForOp, YieldOp
    from xdsl.dialects.math import PowFOp, SqrtOp
    from xdsl.builder import Builder, InsertPoint
    from xdsl.transforms.common_subexpression_elimination import cse
    from xdsl.transforms.dead_code_elimination import dce
    return (
        Abs,
        Add,
        AddfOp,
        AddiOp,
        Attribute,
        Block,
        Builder,
        ConstantOp,
        DivfOp,
        E,
        Expr,
        Float,
        Float64Type,
        FloatAttr,
        FloatingPointLikeBinaryOperation,
        ForOp,
        FuncOp,
        Function,
        GreedyRewritePatternApplier,
        I,
        IRDLOperation,
        InsertPoint,
        Integer,
        IntegerAttr,
        IntegerType,
        ModuleOp,
        Mul,
        MulfOp,
        MuliOp,
        Operation,
        ParameterDef,
        ParametrizedAttribute,
        PatternRewriteWalker,
        PatternRewriter,
        Pow,
        PowFOp,
        Pure,
        Rational,
        Region,
        ReturnOp,
        RewritePattern,
        S,
        SIToFPOp,
        SSAValue,
        SqrtOp,
        SubfOp,
        Sum,
        Symbol,
        TypeAttribute,
        YieldOp,
        cse,
        dce,
        im,
        irdl_attr_definition,
        irdl_op_definition,
        mo,
        op_type_rewrite_pattern,
        operand_def,
        re,
        region_dce,
        result_def,
        symbols,
        traits_def,
    )


@app.cell(hide_code=True)
def _(Function):
    Norm = Function("Norm", real=True)
    return (Norm,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        # Exercise: Add a `complex` dialect

        Your task is to:

        * Write a new set of `complex` operations
        * Add a lowering from `complex` to `arith`
        * Add optimizations for the `complex` dialect
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        ## Defining the `complex` dialect

        You should write the following operations:

        * "complex.create": creates a complex value given its real and imaginary float types
        * "complex.re": returns the real part of a complex number
        * "complex.im": returns the imaginary part of a complex number
        * "complex.add": adds two complex numbers
        * "complex,mul": multiples two copmlex numbers
        * "complex.norm": returns the norm of two complex numbers

        Here is the definition of the `!complex.complex` type, and the definition of `complex.re`. Complete it with the other ops:
        """
    )
    return


@app.cell
def _(
    Float64Type,
    IRDLOperation,
    ParametrizedAttribute,
    Pure,
    SSAValue,
    TypeAttribute,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
    traits_def,
):
    @irdl_attr_definition
    class ComplexType(ParametrizedAttribute, TypeAttribute):
        name = "complex.complex"

    @irdl_op_definition
    class ReOp(IRDLOperation):
        name = "complex.re"
        traits = traits_def(Pure())

        arg = operand_def(ComplexType())

        result = result_def(Float64Type())

        def __init__(self, arg: SSAValue):
            super().__init__(operands=[arg], result_types=[Float64Type()])

    @irdl_op_definition
    class ImOp(IRDLOperation):
        name = "complex.im"
        traits = traits_def(Pure())

        # Name the fields arg and result

        def __init__(self, arg: SSAValue):
            raise NotImplementedError("ImOp __init__ is not yet implemented")

    @irdl_op_definition
    class CreateOp(IRDLOperation):
        name = "complex.create"
        traits = traits_def(Pure())

        # Name the fields re, im, and result

        def __init__(self, re: SSAValue, im: SSAValue):
            raise NotImplementedError("CreateOp __init__ is not yet implemented")

    @irdl_op_definition
    class AddcOp(IRDLOperation):
        name = "complex.add"
        traits = traits_def(Pure())

        # Name the fields lhs, rhs, and result

        def __init__(self, lhs: SSAValue, rhs: SSAValue):
            raise NotImplementedError("AddcOp __init__ is not yet implemented")

    @irdl_op_definition
    class MulcOp(IRDLOperation):
        name = "complex.mul"
        traits = traits_def(Pure())

        # Name the fields lhs, rhs, and result

        def __init__(self, lhs: SSAValue, rhs: SSAValue):
            raise NotImplementedError("MulcOp __init__ is not yet implemented")

    @irdl_op_definition
    class NormOp(IRDLOperation):
        name = "complex.norm"
        traits = traits_def(Pure())

        # Name the fields arg and result

        def __init__(self, arg: SSAValue):
            raise NotImplementedError("NormOp __init__ is not yet implemented")
    return AddcOp, ComplexType, CreateOp, ImOp, MulcOp, NormOp, ReOp


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        ### Solution

        Hidden below is the definition of all operations
        """
    )
    return


@app.cell(hide_code=True)
def _(
    Float64Type,
    IRDLOperation,
    ParametrizedAttribute,
    Pure,
    SSAValue,
    TypeAttribute,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
    traits_def,
):
    def dialect_solution():
        @irdl_attr_definition
        class ComplexType(ParametrizedAttribute, TypeAttribute):
            name = "complex.complex"

        @irdl_op_definition
        class CreateOp(IRDLOperation):
            name = "complex.create"

            traits = traits_def(Pure())

            re = operand_def(Float64Type())
            im = operand_def(Float64Type())

            result = result_def(ComplexType())

            def __init__(self, re: SSAValue, im: SSAValue):
                super().__init__(operands=[re, im], result_types=[ComplexType()])

        @irdl_op_definition
        class ReOp(IRDLOperation):
            name = "complex.re"

            traits = traits_def(Pure())

            arg = operand_def(ComplexType())

            result = result_def(Float64Type())

            def __init__(self, arg: SSAValue):
                super().__init__(operands=[arg], result_types=[Float64Type()])

        @irdl_op_definition
        class ImOp(IRDLOperation):
            name = "complex.im"

            traits = traits_def(Pure())

            arg = operand_def(ComplexType())

            result = result_def(Float64Type())

            def __init__(self, arg: SSAValue):
                super().__init__(operands=[arg], result_types=[Float64Type()])

        @irdl_op_definition
        class AddcOp(IRDLOperation):
            name = "complex.add"

            traits = traits_def(Pure())

            lhs = operand_def(ComplexType())
            rhs = operand_def(ComplexType())

            result = result_def(ComplexType())

            def __init__(self, lhs: SSAValue, rhs: SSAValue):
                super().__init__(operands=[lhs, rhs], result_types=[ComplexType()])

        @irdl_op_definition
        class MulcOp(IRDLOperation):
            name = "complex.mul"

            traits = traits_def(Pure())

            lhs = operand_def(ComplexType())
            rhs = operand_def(ComplexType())

            result = result_def(ComplexType())

            def __init__(self, lhs: SSAValue, rhs: SSAValue):
                super().__init__(operands=[lhs, rhs], result_types=[ComplexType()])

        @irdl_op_definition
        class NormOp(IRDLOperation):
            name = "complex.norm"

            traits = traits_def(Pure())

            arg = operand_def(ComplexType())

            result = result_def(Float64Type())

            def __init__(self, arg: SSAValue):
                super().__init__(operands=[arg], result_types=[Float64Type()])
    return (dialect_solution,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Filling the IR emitter with `complex` operations:

        Now that our new operations are defined, we can add support for them in the IR emitter.
        We completed this for you below.
        """
    )
    return


@app.cell(hide_code=True)
def _(
    Builder,
    Expr,
    FuncOp,
    InsertPoint,
    ModuleOp,
    ReturnOp,
    cse,
    emit_op,
    get_mlir_type,
):
    def emit_ir(expr: Expr) -> ModuleOp:
        # Create a module, and create a builder at the beginning of its only block
        module = ModuleOp([])
        builder = Builder(InsertPoint.at_end(module.body.block))

        # Create the MLIR types for each symbol.
        arg_types = [get_mlir_type(arg) for arg in expr.free_symbols]

        # Create a new function and inserts it inside the module.
        func = FuncOp("main", (arg_types, [get_mlir_type(expr)]))
        builder.insert(func)

        # Associate each symbol with its MLIR name.
        arg_values = {arg: value for arg, value in zip(expr.free_symbols, func.args)}

        # Set the name for each function argument. This is only to get better names
        # for IR values.
        for arg, value in arg_values.items():
            value.name_hint = arg.name

        # Set the builder insertion point inside the function.
        builder.insertion_point = InsertPoint.at_end(func.body.block)

        # Convert the expression into MLIR IR inside the function.
        result = emit_op(expr, builder, arg_values)

        # Insert a return statement at the end of the function.
        builder.insert(ReturnOp(result))
        return module

    def print_ir(expr: Expr):
        try:
            # Print the SymPy expression
            print(expr)

            # Converts the SymPy expression to an MLIR `builtin.module` operation
            op = emit_ir(expr)
            cse(op)

            # Check that the operation verifies, and prints the operation
            op.verify()
            print(op)

            # Print a separator
            print("\n\n")
        except NotImplementedError as e:
            print("Error:", e)
    return emit_ir, print_ir


@app.cell(hide_code=True)
def _(Attribute, ComplexType, Expr, Float64Type, IntegerType):
    def get_mlir_type(expr: Expr) -> Attribute:
        if expr.is_integer:
            return IntegerType(64)
        elif expr.is_real:
            return Float64Type()
        else:
            return ComplexType()
    return (get_mlir_type,)


@app.cell(hide_code=True)
def _(
    Abs,
    Add,
    AddcOp,
    AddfOp,
    AddiOp,
    Block,
    Builder,
    CmpfOp,
    ComplexType,
    ConstantOp,
    CreateOp,
    Expr,
    Float,
    Float64Type,
    FloatAttr,
    ForOp,
    I,
    IfOp,
    InsertPoint,
    Integer,
    IntegerAttr,
    IntegerType,
    Mul,
    MulcOp,
    MulfOp,
    MuliOp,
    Norm,
    NormOp,
    Pow,
    PowFOp,
    Region,
    SIToFPOp,
    SSAValue,
    SelectOp,
    SubfOp,
    Sum,
    Symbol,
    YieldOp,
    get_mlir_type,
):
    def emit_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        type = get_mlir_type(expr)
        if isinstance(type, IntegerType):
            return emit_integer_op(expr, builder, args)
        elif isinstance(type, Float64Type):
            return emit_real_op(expr, builder, args)
        elif isinstance(type, ComplexType):
            return emit_complex_op(expr, builder, args)

        raise Exception("Unknown function to emit IR for MLIR type ", type)

    def emit_integer_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        # Handle symbolic values
        if isinstance(expr, Symbol):
            return args[expr]

        # Handle constants
        if isinstance(expr, Integer):
            constant_op = builder.insert(
                ConstantOp(IntegerAttr(int(expr), IntegerType(64)))
            )
            return constant_op.result

        if isinstance(expr, Add):
            lhs = emit_integer_op(expr.args[0], builder, args)
            rhs = emit_integer_op(expr.args[1], builder, args)
            add_op = builder.insert(AddiOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Mul):
            lhs = emit_integer_op(expr.args[0], builder, args)
            rhs = emit_integer_op(expr.args[1], builder, args)
            add_op = builder.insert(MuliOp(lhs, rhs))
            return add_op.result

        raise NotImplementedError(f"No IR emitter for integer function {expr.func}")

    def emit_complex_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        if isinstance(get_mlir_type(expr), Float64Type):
            res = emit_real_op(expr, builder, args)
            zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
            create = builder.insert(CreateOp(res, zero)).result
            return create

        if expr == I:
            zero = builder.insert(ConstantOp(FloatAttr(float(0), Float64Type()))).result
            one = builder.insert(ConstantOp(FloatAttr(float(1), Float64Type()))).result
            create = builder.insert(CreateOp(zero, one)).result
            return create

        if isinstance(expr, Add):
            lhs = emit_complex_op(expr.args[0], builder, args)
            rhs = emit_complex_op(expr.args[1], builder, args)
            res = builder.insert(AddcOp(lhs, rhs)).result
            return res

        if isinstance(expr, Mul):
            lhs = emit_complex_op(expr.args[0], builder, args)
            rhs = emit_complex_op(expr.args[1], builder, args)
            res = builder.insert(MulcOp(lhs, rhs)).result
            return res

        raise NotImplementedError("Not implemented")

    def emit_real_op(
        expr: Expr,
        builder: Builder,
        args: dict[Symbol, SSAValue],
    ):
        # If the expression is an integer expression, emits it and then convert it
        # back to a float expression.
        if expr.is_integer:
            res = emit_integer_op(expr, builder, args)
            op = builder.insert(SIToFPOp(res, Float64Type()))
            return op.result

        # Handle constants
        if isinstance(expr, Float):
            constant_op = builder.insert(
                ConstantOp(FloatAttr(float(expr), Float64Type()))
            )
            return constant_op

        # Handle symbolic values
        if isinstance(expr, Symbol):
            return args[expr]

        if isinstance(expr, Add):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(AddfOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Mul):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(MulfOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Pow):
            lhs = emit_real_op(expr.args[0], builder, args)
            rhs = emit_real_op(expr.args[1], builder, args)
            add_op = builder.insert(PowFOp(lhs, rhs))
            return add_op.result

        if isinstance(expr, Abs):
            # The arith.select solution
            if False:
                arg = emit_real_op(expr.args[0], builder, args)
                zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
                neg = builder.insert(SubfOp(zero, arg)).result
                is_neg = builder.insert(CmpfOp(arg, zero, "olt")).result
                select = builder.insert(SelectOp(is_neg, neg, arg)).result
                return select

            # The scf.if solution
            arg = emit_real_op(expr.args[0], builder, args)
            zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
            is_neg = builder.insert(CmpfOp(arg, zero, "olt")).result

            lhs_region = Region([Block()])
            builder2 = Builder(InsertPoint.at_end(lhs_region.block))
            neg = builder2.insert(SubfOp(zero, arg)).result
            builder2.insert(YieldOp(neg))

            rhs_region = Region([Block()])
            builder3 = Builder(InsertPoint.at_end(rhs_region.block))
            builder3.insert(YieldOp(neg))

            if_res = builder.insert(
                IfOp(is_neg, Float64Type(), lhs_region, rhs_region)
            ).results[0]

            return if_res

        if isinstance(expr, Sum):
            zero = builder.insert(ConstantOp(FloatAttr(0, Float64Type()))).result
            lb = emit_integer_op(expr.args[1][1], builder, args)
            ub = emit_integer_op(expr.args[1][2], builder, args)
            step = builder.insert(ConstantOp(IntegerAttr(0, IntegerType(64))))
            region = Region([Block(arg_types=[IntegerType(64), Float64Type()])])
            accumulator = region.block.args[1]

            b2 = Builder(InsertPoint.at_end(region.block))
            arg = emit_real_op(
                expr.args[0], b2, args | {expr.args[1][0]: region.block.args[0]}
            )
            add = b2.insert(AddfOp(arg, accumulator)).result
            b2.insert(YieldOp(add))

            return builder.insert(ForOp(lb, ub, step, [zero], region)).results[0]

        if isinstance(expr, Norm):
            arg = emit_complex_op(expr.args[0], builder, args)
            res = builder.insert(NormOp(arg)).result
            return res

        raise NotImplementedError(f"No IR emitter for float function {expr.func}")
    return emit_complex_op, emit_integer_op, emit_op, emit_real_op


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Here are a few examples of the code that we can generate:""")
    return


@app.cell
def _(symbols):
    x, y, z, t = symbols("x y z t", real=True)
    return t, x, y, z


@app.cell
def _(I, print_ir, x, y):
    print_ir(x + I * y)
    return


@app.cell
def _(I, print_ir, t, x, y, z):
    print_ir((x + I * y) * (z + I * t))
    return


@app.cell
def _(I, Norm, print_ir, x, y):
    print_ir(Norm(x + I * y))
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        ## Lowering the complex dialect

        Now that we emitted IR using the `complex` dialect, we can lower it to `arith`.

        To do this, follow these steps:

        * Write the patterns `re(create(x, y)) -> x` and `im(create(x, y))`
        * Write patterns to rewrite `complex.add`, `complex.mul`, and `complex.norm` into `arith` and `complex.create`, `complex.re`, and `complex.im` ops. For instance, `add(x, y)` should be rewritten to `create(re(x) + re(y), im(x) + im(y))`.

        This will effectively lower the `complex` dialect, as all these patterns will be applied until convergence. We give you the pattern `re(create(x, y)) -> x` and the lowering of `complex.mul`
        """
    )
    return


@app.cell
def _(
    AddfOp,
    CreateOp,
    GreedyRewritePatternApplier,
    ImOp,
    MulcOp,
    MulfOp,
    Operation,
    PatternRewriteWalker,
    PatternRewriter,
    ReOp,
    RewritePattern,
    SubfOp,
    cse,
    dce,
):
    # Lower the complex dialect.
    def lower_complex(op: Operation):
        # Hint: Add rewrite patterns in this list
        rewrites = [
            FoldReCreateOp(),
            FoldImCreateOp(),
            LowerAddOp(),
            LowerMulOp(),
            LowerNormOp(),
        ]
        PatternRewriteWalker(GreedyRewritePatternApplier(rewrites)).rewrite_module(op)

        # Run dce and cse after the rewritting
        cse(op)
        dce(op)

    class FoldReCreateOp(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if not isinstance(op, ReOp):
                return

            if not isinstance(create := op.arg.owner, CreateOp):
                return

            rewriter.replace_matched_op([], new_results=[create.re])

    class FoldImCreateOp(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            # Implement im(create(x, y)) -> y
            return

    class LowerAddOp(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            # Implement the lowering of add
            # The formula is `add(x, y) = (re(x) + re(y)) + i * (im(x) + im(y))
            return

    class LowerMulOp(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if not isinstance(op, MulcOp):
                return

            re_lhs = rewriter.insert(ReOp(op.lhs)).result
            re_rhs = rewriter.insert(ReOp(op.rhs)).result
            im_lhs = rewriter.insert(ImOp(op.lhs)).result
            im_rhs = rewriter.insert(ImOp(op.rhs)).result

            tmp1 = rewriter.insert(MulfOp(re_lhs, re_rhs)).result
            tmp2 = rewriter.insert(MulfOp(im_lhs, im_rhs)).result
            new_re = rewriter.insert(SubfOp(tmp1, tmp2)).result

            tmp3 = rewriter.insert(MulfOp(re_lhs, im_rhs)).result
            tmp4 = rewriter.insert(MulfOp(im_lhs, re_rhs)).result
            new_im = rewriter.insert(AddfOp(tmp3, tmp4)).result

            create = rewriter.insert(CreateOp(new_re, new_im)).result

            rewriter.replace_matched_op([], new_results=[create])

    class LowerNormOp(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            # Implement the lowering of norm
            # The formula is `norm(z) = (re(z) * re(z) + im(z) * im(z)) ^ 0.5`
            return
    return (
        FoldImCreateOp,
        FoldReCreateOp,
        LowerAddOp,
        LowerMulOp,
        LowerNormOp,
        lower_complex,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""You can test your code using the `print_ir_with_complex_lowering` function:""")
    return


@app.cell(hide_code=True)
def _(Expr, cse, emit_ir, lower_complex):
    def print_ir_with_complex_lowering(expr: Expr):
        try:
            # Print the SymPy expression
            print(expr)

            # Converts the SymPy expression to an MLIR `builtin.module` operation
            op = emit_ir(expr)
            cse(op)

            # Check that the operation verifies, and prints the operation
            op.verify()
            print("Op before lowering:")
            print(op)

            print("Op after lowering:")
            lower_complex(op)
            op.verify()
            print(op)

            # Print a separator
            print("\n\n")
        except NotImplementedError as e:
            print("Error:", e)
    return (print_ir_with_complex_lowering,)


@app.cell
def _(I, Norm, print_ir_with_complex_lowering, x, y):
    print_ir_with_complex_lowering(Norm(x + I * y))
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        ### Solution

        Hidden below is a possible lowering:
        """
    )
    return


@app.cell(hide_code=True)
def _(
    AddcOp,
    AddfOp,
    ConstantOp,
    CreateOp,
    Float64Type,
    FloatAttr,
    GreedyRewritePatternApplier,
    ImOp,
    MulcOp,
    MulfOp,
    NormOp,
    Operation,
    PatternRewriteWalker,
    PatternRewriter,
    PowFOp,
    ReOp,
    RewritePattern,
    SubfOp,
    cse,
    dce,
):
    def solution():
        # Lower the complex dialect.
        def lower_complex(op: Operation):
            # Hint: Add rewrite patterns in this list
            rewrites = [
                FoldReCreateOp(),
                FoldImCreateOp(),
                LowerAddOp(),
                LowerMulOp(),
                LowerNormOp(),
            ]
            PatternRewriteWalker(GreedyRewritePatternApplier(rewrites)).rewrite_module(
                op
            )

            # Run dce and cse after the rewritting
            cse(op)
            dce(op)

        class FoldReCreateOp(RewritePattern):
            def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
                if not isinstance(op, ReOp):
                    return

                if not isinstance(create := op.arg.owner, CreateOp):
                    return

                rewriter.replace_matched_op([], new_results=[create.re])

        class FoldImCreateOp(RewritePattern):
            def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
                if not isinstance(op, ImOp):
                    return

                if not isinstance(create := op.arg.owner, CreateOp):
                    return

                rewriter.replace_matched_op([], new_results=[create.im])

        class LowerAddOp(RewritePattern):
            def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
                if not isinstance(op, AddcOp):
                    return

                re_lhs = rewriter.insert(ReOp(op.lhs)).result
                re_rhs = rewriter.insert(ReOp(op.rhs)).result
                im_lhs = rewriter.insert(ImOp(op.lhs)).result
                im_rhs = rewriter.insert(ImOp(op.rhs)).result
                new_re = rewriter.insert(AddfOp(re_lhs, re_rhs)).result
                new_im = rewriter.insert(AddfOp(im_lhs, im_rhs)).result
                create = rewriter.insert(CreateOp(new_re, new_im)).result

                rewriter.replace_matched_op([], new_results=[create])

        class LowerMulOp(RewritePattern):
            def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
                if not isinstance(op, MulcOp):
                    return

                re_lhs = rewriter.insert(ReOp(op.lhs)).result
                re_rhs = rewriter.insert(ReOp(op.rhs)).result
                im_lhs = rewriter.insert(ImOp(op.lhs)).result
                im_rhs = rewriter.insert(ImOp(op.rhs)).result

                tmp1 = rewriter.insert(MulfOp(re_lhs, re_rhs)).result
                tmp2 = rewriter.insert(MulfOp(im_lhs, im_rhs)).result
                new_re = rewriter.insert(SubfOp(tmp1, tmp2)).result

                tmp3 = rewriter.insert(MulfOp(re_lhs, im_rhs)).result
                tmp4 = rewriter.insert(MulfOp(im_lhs, re_rhs)).result
                new_im = rewriter.insert(AddfOp(tmp3, tmp4)).result

                create = rewriter.insert(CreateOp(new_re, new_im)).result

                rewriter.replace_matched_op([], new_results=[create])

        class LowerNormOp(RewritePattern):
            def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
                if not isinstance(op, NormOp):
                    return

                re = rewriter.insert(ReOp(op.arg)).result
                im = rewriter.insert(ImOp(op.arg)).result

                re_2 = rewriter.insert(MulfOp(re, re)).result
                im_2 = rewriter.insert(MulfOp(im, im)).result

                add = rewriter.insert(AddfOp(re_2, im_2)).result

                half = rewriter.insert(ConstantOp(FloatAttr(0.5, Float64Type()))).result
                pow = rewriter.insert(PowFOp(add, half)).result

                rewriter.replace_matched_op([], new_results=[pow])
    return (solution,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
        ## Optimizing our dialects

        As a last task, you have to write your own optimizations patterns to improve as much as possible the performance of the generated code. This includes reducing the amount of floating point power and multiplication functions.

        Use `print_ir_with_pipeline` to test your code, it will run an optimization pass, then the lowering of `complex`, then another optimization pass.

        Your objective is to optimize the function `Norm(z1) * Norm(z2)`.

        As a hint, this expression is equivalent to `Norm(z1 * z2)`, which is much easier to express as a `complex` dialect optimization than an `arith` optimization. You may also add other `arith` and `complex` optimizations.
        """
    )
    return


@app.cell(hide_code=True)
def _(Expr, cse, dce, emit_ir, lower_complex, optimize):
    def print_ir_with_pipeline(expr: Expr):
        try:
            # Print the SymPy expression
            print(expr)

            # Converts the SymPy expression to an MLIR `builtin.module` operation
            print("Emitted IR:")
            op = emit_ir(expr)
            cse(op)
            dce(op)
            op.verify()
            print(op)
            print("\n\n")

            print("After first optimization:")
            optimize(op)
            op.verify()
            print(op)
            print("\n\n")

            print("After complex lowering:")
            lower_complex(op)
            op.verify()
            print(op)
            print("\n\n")

            print("After second optimization:")
            optimize(op)
            op.verify()
            print(op)
            print("\n\n")

            # Print a separator
            print("\n\n")
        except NotImplementedError as e:
            print("Error:", e)
    return (print_ir_with_pipeline,)


@app.cell
def _(GreedyRewritePatternApplier, Operation, PatternRewriteWalker, cse, dce):
    # Optimize the `complex` and `arith` dialects
    def optimize(op: Operation):
        # Hint: Add rewrite patterns in this list
        rewrites = []
        PatternRewriteWalker(GreedyRewritePatternApplier(rewrites)).rewrite_module(op)

        # Run cse and dce
        cse(op)
        dce(op)

    # Define new patterns here
    return (optimize,)


@app.cell
def _(I, Norm, print_ir_with_pipeline, t, x, y, z):
    print_ir_with_pipeline(Norm(x + I * y) * Norm(z + I * t))
    return


if __name__ == "__main__":
    app.run()



import pytest

from xdsl.context import Context
from xdsl.dialects import get_all_dialects
from xdsl.irdl.dominance import strictly_dominates
from xdsl.parser import Parser

ctx = Context()
ctx.register_dialect("test", get_all_dialects()["test"])


# Create a block graph as in https://en.wikipedia.org/w/index.php?title=Dominator_(graph_theory)&oldid=1189814332

op = Parser(
    ctx,
    """
"test.op"() ({
^1():
  "test.op"()[^2] : () -> ()
^2():
  "test.op"()[^3, ^4, ^6] : () -> ()
^3():
  "test.op"()[^5] : () -> ()
^4():
  "test.op"()[^5] : () -> ()
^5():
  "test.op"()[^2] : () -> ()
^6():
  "test.op"() : () -> ()
}) : () -> ()
""",
).parse_op()

blocks = op.regions[0].blocks


@pytest.mark.parametrize(
    ("a", "b", "expected"),
    [
        (1, 1, False),
        (1, 2, True),
        (1, 3, True),
        (1, 4, True),
        (1, 5, True),
        (1, 6, True),
        (2, 2, False),
        (2, 3, True),
        (2, 4, True),
        (2, 5, True),
        (2, 6, True),
        (3, 3, False),
        (3, 4, False),
        (3, 5, False),
        (3, 6, False),
        (4, 4, False),
        (4, 5, False),
        (4, 6, False),
        (5, 5, False),
        (5, 6, False),
        (6, 6, False),
    ],
)
def test_region_strictly_dominates_block(a: int, b: int, expected: bool):
    """
    Test in-region block dominance.
    """
    assert strictly_dominates(blocks[a - 1], blocks[b - 1]) == expected



from collections.abc import Sequence
from typing import cast

from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.dialects.arith import AddiOp, CmpiOp, IndexCastOp
from xdsl.dialects.builtin import DenseIntOrFPElementsAttr, i32
from xdsl.dialects.cf import BranchOp, ConditionalBranchOp, SwitchOp
from xdsl.dialects.scf import ForOp, IfOp, IndexSwitchOp, YieldOp
from xdsl.ir import Block, Region
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import BlockInsertPoint, InsertPoint
from xdsl.traits import IsTerminator


class IfLowering(RewritePattern):
    """
    Lowers `scf.if` to conditional branching.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, if_op: IfOp, rewriter: PatternRewriter, /):
        condition_block = if_op.parent_block()
        assert condition_block is not None

        # Start by splitting the block containing the 'scf.if' into two parts.
        # The part before will contain the condition, the part after will be the
        # continuation point.
        remaining_ops_block = condition_block.split_before(if_op)
        if if_op.results:
            parent = condition_block.parent_region()
            assert parent is not None
            continue_block = Block(arg_types=if_op.result_types)
            parent.insert_block_before(continue_block, remaining_ops_block)
            rewriter.insert_op(
                BranchOp(remaining_ops_block), InsertPoint.at_end(continue_block)
            )
        else:
            continue_block = remaining_ops_block

        # Move blocks from the "then" region to the region containing 'scf.if',
        # place it before the continuation block, and branch to it.
        then_region = if_op.true_region
        then_block = then_region.first_block
        assert then_block is not None
        assert then_region.last_block is not None
        then_terminator = then_region.last_block.last_op
        assert then_terminator is not None
        then_terminator_operands = then_terminator.operands
        rewriter.insert_op(
            BranchOp(continue_block, *then_terminator_operands),
            InsertPoint.at_end(then_region.last_block),
        )

        rewriter.erase_op(then_terminator)
        rewriter.inline_region(then_region, BlockInsertPoint.before(continue_block))

        # Move blocks from the "else" region (if present) to the region containing
        # 'scf.if', place it before the continuation block and branch to it.  It
        # will be placed after the "then" regions.
        else_region = if_op.false_region
        else_block = else_region.first_block
        if else_block is not None:
            assert else_region.last_block is not None
            else_terminator = else_region.last_block.last_op
            assert else_terminator is not None
            else_terminator_operands = else_terminator.operands
            rewriter.insert_op(
                BranchOp(continue_block, *else_terminator_operands),
                InsertPoint.at_end(else_region.last_block),
            )

            rewriter.erase_op(else_terminator)
            rewriter.inline_region(else_region, BlockInsertPoint.before(continue_block))
        else:
            else_block = continue_block

        # Branch to either the then_block or else_block
        rewriter.insert_op(
            ConditionalBranchOp(if_op.cond, then_block, (), else_block, ()),
            InsertPoint.at_end(condition_block),
        )

        # Remove the original `scf.if` operation
        rewriter.replace_matched_op([], continue_block.args)


class ForLowering(RewritePattern):
    """Lowers `scf.for` to conditional branching."""

    @op_type_rewrite_pattern
    def match_and_rewrite(self, for_op: ForOp, rewriter: PatternRewriter):
        # Start by splitting the block containing the 'scf.for' into two parts.
        # The part before will get the init code, the part after will be the end
        # point.
        init_block = for_op.parent_block()
        if init_block is None:
            return
        end_block = init_block.split_before(for_op)

        # Use the first block of the loop body as the condition block since it is the
        # block that has the induction variable and loop-carried values as arguments.
        # Split out all operations from the first block into a new block. Move all
        # body blocks from the loop body region to the region containing the loop.
        condition_block = for_op.body.first_block
        assert condition_block is not None
        first_op = condition_block.first_op
        assert first_op is not None
        first_body_block = condition_block.split_before(first_op)
        last_body_block = for_op.body.last_block
        assert last_body_block is not None
        rewriter.inline_region(for_op.body, BlockInsertPoint.before(end_block))
        iv = condition_block.args[0]

        # Append the induction variable stepping logic to the last body block and
        # branch back to the condition block. Loop-carried values are taken from
        # operands of the loop terminator.
        terminator = last_body_block.last_op
        assert terminator is not None
        assert terminator.has_trait(IsTerminator)

        stepped = AddiOp(iv, for_op.step)
        rewriter.insert_op(stepped, InsertPoint.before(terminator))

        rewriter.replace_op(
            terminator, BranchOp(condition_block, stepped, *terminator.operands)
        )

        # The initial values of loop-carried values are obtained from the operands
        # of the loop operation.
        rewriter.insert_op(
            BranchOp(condition_block, for_op.lb, *for_op.iter_args),
            InsertPoint.at_end(init_block),
        )

        # With the body block done, we can fill in the condition block.
        comparison = CmpiOp(iv, for_op.ub, "slt")
        rewriter.insert_op(comparison, InsertPoint.at_end(condition_block))
        cond_branch_op = ConditionalBranchOp(
            comparison, first_body_block, (), end_block, ()
        )
        rewriter.insert_op(cond_branch_op, InsertPoint.at_end(condition_block))

        # The result of the loop operation are the values of the condition block
        # arguments except the induction variable on the last iteration.
        rewriter.replace_matched_op([], condition_block.args[1:])


class SwitchLowering(RewritePattern):
    """Lowers `scf.index_switch` to `cf.switch`."""

    @staticmethod
    def _convert_region(
        region: Region, continue_block: Block, rewriter: PatternRewriter
    ) -> Block:
        block = region.first_block
        assert block is not None

        # Convert yield op to a branch to the continue block
        yield_op = block.last_op
        assert isinstance(yield_op, YieldOp)
        rewriter.replace_op(yield_op, BranchOp(continue_block, *yield_op.operands))

        # Inline the region
        rewriter.inline_region(region, BlockInsertPoint.before(continue_block))
        return block

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: IndexSwitchOp, rewriter: PatternRewriter):
        # Split the block at `op`
        condition_block = op.parent_block()
        if condition_block is None:
            return

        continue_block = condition_block.split_before(op)

        # Create the arguments on the continue block with which to replace the
        # results of the op.
        for i, ty in enumerate(op.result_types):
            rewriter.insert_block_argument(continue_block, i, ty)

        # Convert the case regions
        case_successors = tuple(
            self._convert_region(region, continue_block, rewriter)
            for region in op.case_regions
        )
        case_values = cast(Sequence[int], op.cases.get_values())

        # Convert the default region
        default_block = self._convert_region(
            op.default_region, continue_block, rewriter
        )

        # Cast switch index to integer case value
        case_value = IndexCastOp(op.arg, i32)
        rewriter.insert_op(case_value, InsertPoint.at_end(condition_block))

        # Create the switch
        case_operands = tuple(() for _ in case_successors)
        rewriter.insert_op(
            SwitchOp(
                case_value,
                default_block,
                (),
                DenseIntOrFPElementsAttr.vector_from_list(case_values, i32),
                case_successors,
                case_operands,
            ),
            InsertPoint.at_end(condition_block),
        )

        rewriter.replace_matched_op((), continue_block.args)


class ConvertScfToCf(ModulePass):
    """
    Lower `scf.for` and `scf.if` to unstructured control flow.
    Implementations are direct translations of the mlir versions found at
    https://github.com/llvm/llvm-project/blob/main/mlir/lib/Conversion/SCFToControlFlow/SCFToControlFlow.cpp
    """

    name = "convert-scf-to-cf"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    SwitchLowering(),
                    IfLowering(),
                    ForLowering(),
                ]
            )
        ).rewrite_module(op)



import ast
from typing import Generic, TypeVar

_T = TypeVar("_T")


class Const(Generic[_T]):
    """
    Marks the value as a compile-time constant and can be used for any type
    defined by the frontend. Compile-time constants can be expressions and are
    evaluated at compile time. They cannot be redefined in the same program or
    shadowed by other variables.

    ```
    a: Const[i64] = 123             # i64 constant equal to 123
    b: Const[i32] = len([1, 2, 3])  # i32 constant equal to 3

    c: Const[f32] = 123 / 0         # compile-time error, division by zero

    d: Const[i64] = a + b           # i64 constant, equal to 6

    def foo(a: i64):                # compile-time error, cannot reuse the name
                                    # of a constant

        d = 3                       # compile-time error, cannot assign to a constant
        e: Const[i16] = b + 2       # i16 constant, equal to 5
    ```
    """

    pass


def is_constant(node: ast.expr) -> bool:
    """Returns `True` if the AST node is a Const type."""
    return (
        isinstance(node, ast.Subscript)
        and isinstance(node.value, ast.Name)
        and node.value.id == Const.__name__
    )


def is_constant_stmt(node: ast.stmt) -> bool:
    """Returns `True` if the AST statement is a Const expression."""
    return isinstance(node, ast.AnnAssign) and is_constant(node.annotation)



from collections.abc import Iterable
from dataclasses import dataclass
from typing import cast

from xdsl.context import Context
from xdsl.dialects import arith, builtin, func, memref, ptr
from xdsl.ir import Attribute, Operation, SSAValue
from xdsl.irdl import Any
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.exceptions import DiagnosticException


def offset_calculations(
    memref_type: memref.MemRefType[Any], indices: Iterable[SSAValue]
) -> tuple[list[Operation], SSAValue]:
    """
    Get operations calculating an offset which needs to be added to memref's base
    pointer to access an element referenced by indices.
    """

    assert isinstance(memref_type.element_type, builtin.FixedBitwidthType)

    match memref_type.layout:
        case builtin.NoneAttr():
            strides = builtin.ShapedType.strides_for_shape(memref_type.get_shape())
        case builtin.StridedLayoutAttr():
            strides = memref_type.layout.get_strides()
        case _:
            raise DiagnosticException(f"Unsupported layout type {memref_type.layout}")

    ops: list[Operation] = []

    head: SSAValue | None = None

    for index, stride in zip(indices, strides, strict=True):
        # Calculate the offset that needs to be added through the index of the current
        # dimension.
        increment = index
        match stride:
            case None:
                raise DiagnosticException(
                    f"MemRef {memref_type} with dynamic stride is not yet implemented"
                )
            case 1:
                # Stride 1 is a noop making the index equal to the offset.
                pass
            case _:
                # Otherwise, multiply the stride (which by definition is the number of
                # elements required to be skipped when incrementing that dimension).
                ops.extend(
                    (
                        stride_op := arith.ConstantOp.from_int_and_width(
                            stride, builtin.IndexType()
                        ),
                        offset_op := arith.MuliOp(increment, stride_op),
                    )
                )
                stride_op.result.name_hint = "pointer_dim_stride"
                offset_op.result.name_hint = "pointer_dim_offset"

                increment = offset_op.result

        if head is None:
            # First iteration.
            head = increment
            continue

        # Otherwise sum up the products.
        add_op = arith.AddiOp(head, increment)
        add_op.result.name_hint = "pointer_dim_stride"
        ops.append(add_op)
        head = add_op.result

    if head is None:
        raise DiagnosticException("Got empty indices for offset calculations.")

    ops.extend(
        [
            bytes_per_element_op := ptr.TypeOffsetOp(
                operands=[],
                result_types=[builtin.IndexType()],
                properties={"elem_type": memref_type.element_type},
            ),
            final_offset := arith.MuliOp(head, bytes_per_element_op),
        ]
    )

    bytes_per_element_op.offset.name_hint = "bytes_per_element"
    final_offset.result.name_hint = "scaled_pointer_offset"

    return ops, final_offset.result


def get_target_ptr(
    target_memref: SSAValue,
    memref_type: memref.MemRefType[Any],
    indices: Iterable[SSAValue],
) -> tuple[list[Operation], SSAValue]:
    """Get operations returning a pointer to an element of a memref referenced by indices."""

    ops: list[Operation] = [
        memref_ptr := ptr.ToPtrOp(
            operands=[target_memref], result_types=[ptr.PtrType()]
        )
    ]

    if not indices:
        return ops, memref_ptr.res

    offset_ops, offset = offset_calculations(memref_type, indices)
    ops = offset_ops + ops
    ops.append(
        target_ptr := ptr.PtrAddOp(
            operands=[memref_ptr, offset], result_types=[ptr.PtrType()]
        )
    )

    target_ptr.result.name_hint = "offset_pointer"
    return ops, target_ptr.result


@dataclass
class ConvertStoreOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: memref.StoreOp, rewriter: PatternRewriter, /):
        assert isinstance(op_memref_type := op.memref.type, memref.MemRefType)
        memref_type = cast(memref.MemRefType[Any], op_memref_type)

        ops, target_ptr = get_target_ptr(op.memref, memref_type, op.indices)
        ops.append(ptr.StoreOp(operands=[target_ptr, op.value]))

        rewriter.replace_matched_op(ops)


@dataclass
class ConvertLoadOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: memref.LoadOp, rewriter: PatternRewriter, /):
        assert isinstance(op_memref_type := op.memref.type, memref.MemRefType)
        memref_type = cast(memref.MemRefType[Any], op_memref_type)

        ops, target_ptr = get_target_ptr(op.memref, memref_type, op.indices)
        ops.append(
            load_result := ptr.LoadOp(
                operands=[target_ptr], result_types=[memref_type.element_type]
            )
        )

        rewriter.replace_matched_op(ops, new_results=[load_result.res])


@dataclass
class LowerMemRefFuncOpPattern(RewritePattern):
    """
    Rewrites function arguments of MemRefType to PtrType.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter, /):
        # rewrite function declaration
        new_input_types = [
            ptr.PtrType() if isinstance(arg, builtin.MemRefType) else arg
            for arg in op.function_type.inputs
        ]
        new_output_types = [
            ptr.PtrType() if isinstance(arg, builtin.MemRefType) else arg
            for arg in op.function_type.outputs
        ]
        op.function_type = func.FunctionType.from_lists(
            new_input_types,
            new_output_types,
        )

        if op.is_declaration:
            return

        insert_point = InsertPoint.at_start(op.body.blocks[0])

        # rewrite arguments
        for arg in op.args:
            if not isinstance(arg_type := arg.type, memref.MemRefType):
                continue

            old_type = cast(memref.MemRefType[Attribute], arg_type)
            arg.type = ptr.PtrType()

            if not arg.uses:
                continue

            rewriter.insert_op(
                cast_op := builtin.UnrealizedConversionCastOp.get([arg], [old_type]),
                insert_point,
            )
            arg.replace_by_if(cast_op.results[0], lambda x: x.operation is not cast_op)


@dataclass
class LowerMemRefFuncReturnPattern(RewritePattern):
    """
    Rewrites all `memref` arguments to `func.return` into `ptr.PtrType`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.ReturnOp, rewriter: PatternRewriter, /):
        if not any(isinstance(arg.type, memref.MemRefType) for arg in op.arguments):
            return

        insert_point = InsertPoint.before(op)
        new_arguments: list[SSAValue] = []

        # insert `memref -> ptr` casts for memref return values
        for argument in op.arguments:
            if isinstance(argument.type, memref.MemRefType):
                rewriter.insert_op(
                    cast_op := builtin.UnrealizedConversionCastOp.get(
                        [argument], [ptr.PtrType()]
                    ),
                    insert_point,
                )
                new_arguments.append(cast_op.results[0])
            else:
                new_arguments.append(argument)

        rewriter.replace_matched_op(func.ReturnOp(*new_arguments))


@dataclass
class LowerMemRefFuncCallPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.CallOp, rewriter: PatternRewriter, /):
        if not any(
            isinstance(arg.type, memref.MemRefType) for arg in op.arguments
        ) and not any(isinstance(type, memref.MemRefType) for type in op.result_types):
            return

        # rewrite arguments
        insert_point = InsertPoint.before(op)
        new_arguments: list[SSAValue] = []

        # insert `memref -> ptr` casts for memref arguments values
        for argument in op.arguments:
            if isinstance(argument.type, memref.MemRefType):
                rewriter.insert_op(
                    cast_op := builtin.UnrealizedConversionCastOp.get(
                        [argument], [ptr.PtrType()]
                    ),
                    insert_point,
                )
                new_arguments.append(cast_op.results[0])
            else:
                new_arguments.append(argument)

        insert_point = InsertPoint.after(op)
        new_results: list[SSAValue] = []

        #  insert `ptr -> memref` casts for return values
        for result in op.results:
            if isinstance(result.type, memref.MemRefType):
                rewriter.insert_op(
                    cast_op := builtin.UnrealizedConversionCastOp.get(
                        [result],
                        # TODO: annoying pyright warnings - Sasha, pls help
                        [result.type],  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
                    ),
                    insert_point,
                )
                new_results.append(cast_op.results[0])
            else:
                new_results.append(result)

        new_return_types = [
            ptr.PtrType() if isinstance(type, memref.MemRefType) else type
            for type in op.result_types
        ]

        rewriter.replace_matched_op(
            func.CallOp(op.callee, new_arguments, new_return_types)
        )


class ReconcileUnrealizedPtrCasts(RewritePattern):
    """
    Eliminates two variants of unrealized ptr casts:
    - `ptr_xdsl.ptr -> memref.MemRef -> ptr_xdsl.ptr`;
    - `ptr_xdsl.ptr -> memref.memref` where all uses are `ToPtrOp` operations.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: builtin.UnrealizedConversionCastOp, rewriter: PatternRewriter, /
    ):
        # preconditions
        if (
            len(op.inputs) != 1
            or len(op.outputs) != 1
            or not isinstance(op.inputs[0].type, ptr.PtrType)
            or not isinstance(op.outputs[0].type, memref.MemRefType)
        ):
            return

        # erase ptr -> memref -> ptr cast pairs
        uses = tuple(use for use in op.outputs[0].uses)
        for use in uses:
            if (
                isinstance(use.operation, builtin.UnrealizedConversionCastOp)
                and isinstance(use.operation.inputs[0].type, memref.MemRefType)
                and isinstance(use.operation.outputs[0].type, ptr.PtrType)
            ):
                use.operation.outputs[0].replace_by(op.inputs[0])
                rewriter.erase_op(use.operation)

        # erase this cast entirely if all remaining uses are by ToPtr operations
        cast_ops = [use.operation for use in op.outputs[0].uses]
        if not all(isinstance(op, ptr.ToPtrOp) for op in cast_ops):
            return

        for cast_op in cast_ops:
            cast_op.results[0].replace_by(op.inputs[0])
            rewriter.erase_op(cast_op)

        rewriter.erase_op(op)


@dataclass(frozen=True)
class ConvertMemRefToPtr(ModulePass):
    name = "convert-memref-to-ptr"

    lower_func: bool = False

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier([ConvertStoreOp(), ConvertLoadOp()])
        ).rewrite_module(op)

        if self.lower_func:
            PatternRewriteWalker(
                GreedyRewritePatternApplier(
                    [
                        LowerMemRefFuncOpPattern(),
                        LowerMemRefFuncCallPattern(),
                        LowerMemRefFuncReturnPattern(),
                        ReconcileUnrealizedPtrCasts(),
                    ]
                )
            ).rewrite_module(op)



from xdsl.dialects import arm, arm_func
from xdsl.ir import Region
from xdsl.traits import CallableOpInterface


def test_callable_interface():
    a0, a1 = arm.register.X0, arm.register.X1

    region = Region()
    func = arm_func.FuncOp("callable", region, ((a0, a1), (a0, a1)))

    trait = func.get_trait(CallableOpInterface)

    assert trait is not None

    assert trait.get_callable_region(func) is region
    assert trait.get_argument_types(func) == (a0, a1)
    assert func.assembly_line() is None



from xdsl.dialects import tensor
from xdsl.dialects.csl import csl_stencil
from xdsl.ir import OpResult
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)


class RedundantAccumulatorInitialisation(RewritePattern):
    """
    Removes redundant allocations of empty tensors with no uses other than passed
    as `iter_arg` to `csl_stencil.apply`. Prefer re-use where possible.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: csl_stencil.ApplyOp, rewriter: PatternRewriter
    ) -> None:
        if len(op.accumulator.uses) > 1:
            return

        next_apply = op
        while (next_apply := next_apply.next_op) is not None:
            if (
                isinstance(next_apply, csl_stencil.ApplyOp)
                and len(next_apply.accumulator.uses) == 1
                and isinstance(next_apply.accumulator, OpResult)
                and isinstance(next_apply.accumulator.op, tensor.EmptyOp)
                and op.accumulator.type == next_apply.accumulator.type
            ):
                rewriter.replace_op(next_apply.accumulator.op, [], [op.accumulator])



from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    FloatAttr,
    IntegerAttr,
    MemRefType,
    ModuleOp,
    TensorType,
    f32,
    i32,
    i64,
)
from xdsl.interpreter import Interpreter
from xdsl.interpreters.builtin import BuiltinFunctions
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils import ptr

interpreter = Interpreter(ModuleOp([]))
interpreter.register_implementations(BuiltinFunctions())


def test_values():
    assert interpreter.value_for_attribute(IntegerAttr(1, i32), i32) == 1
    assert interpreter.value_for_attribute(IntegerAttr(2, i64), i64) == 2

    assert interpreter.value_for_attribute(FloatAttr(3.0, f32), f32) == 3.0

    assert interpreter.value_for_attribute(
        DenseIntOrFPElementsAttr.create_dense_int(
            TensorType(i32, [2, 3]), list(range(6))
        ),
        MemRefType(i32, (2, 3)),
    ) == ShapedArray(ptr.TypedPtr.new_int32(list(range(6))), [2, 3])
